import {
  IdentityPoller
} from "./chunk-AZPZPJ27.js";
import {
  convertKeypairToHex,
  createKeypairEDDSA
} from "./chunk-HYTAWJYS.js";
import {
  useStorage
} from "./chunk-7USXUEQQ.js";
import {
  useVisibilityDetection
} from "./chunk-XP2UHWXS.js";
import {
  WebStorage
} from "./chunk-YWADTQF4.js";
import {
  signFrameAction
} from "./chunk-6RZ67RSX.js";

// src/identity/farcaster/use-farcaster-multi-identity.tsx
import {
  useCallback,
  useEffect,
  useMemo,
  useRef,
  useState
} from "react";
var identityReducer = (state, action) => {
  switch (action.type) {
    case "IMPERSONATE": {
      const identity = {
        status: "impersonating",
        fid: action.fid,
        privateKey: action.privateKey,
        publicKey: action.publicKey,
        _id: action.id
      };
      return {
        activeIdentity: identity,
        identities: state.identities.concat(identity)
      };
    }
    case "SELECT_IDENTITY": {
      const activeIdentity = state.identities.find((identity) => identity._id === action.id) ?? null;
      return {
        ...state,
        activeIdentity
      };
    }
    case "LOGOUT": {
      return {
        activeIdentity: null,
        identities: state.identities
      };
    }
    case "REMOVE": {
      const { activeIdentity } = state;
      if (!activeIdentity) {
        return state;
      }
      return {
        activeIdentity: null,
        identities: state.identities.filter(
          (identity) => identity._id !== activeIdentity._id
        )
      };
    }
    case "LOGIN_START": {
      const identity = {
        _id: action.id,
        status: "pending_approval",
        privateKey: action.privateKey,
        publicKey: action.publicKey,
        token: action.token,
        deadline: action.deadline,
        signerApprovalUrl: action.signerApprovalUrl,
        requestFid: action.requestFid,
        requestSigner: action.requestSigner,
        signature: action.signature
      };
      const identities = [
        ...state.identities.filter(
          (storedIdentity) => storedIdentity.status !== "pending_approval"
        ),
        // Add new pending approval
        identity
      ];
      return {
        activeIdentity: identity,
        identities
      };
    }
    case "LOGIN_SUCCESS": {
      const { activeIdentity } = state;
      if (activeIdentity?.status !== "pending_approval") {
        console.warn(
          "Active identity must be selected and be in pending_approval status to be approved"
        );
        return state;
      }
      const updatedActiveIdentity = {
        ...activeIdentity,
        status: "approved",
        fid: action.signedKeyRequest.userFid,
        signedKeyRequest: action.signedKeyRequest
      };
      return {
        activeIdentity: updatedActiveIdentity,
        identities: state.identities.map((identity) => {
          if (identity._id === activeIdentity._id) {
            return updatedActiveIdentity;
          }
          return identity;
        })
      };
    }
    default:
      return state;
  }
};
var defaultStorage = new WebStorage();
var defaultGenerateUserId = () => Date.now();
function useFarcasterMultiIdentity({
  onMissingIdentity,
  signerUrl = "/signer",
  storage = defaultStorage,
  storageKey = "farcasterIdentities",
  visibilityChangeDetectionHook = useVisibilityDetection,
  onImpersonate,
  onLogIn,
  onLogInStart,
  onLogOut,
  onIdentityRemove,
  onIdentitySelect,
  generateUserId = defaultGenerateUserId
}) {
  const storageRef = useRef(storage);
  const identityPoller = useRef(new IdentityPoller()).current;
  const [isLoading, setIsLoading] = useState(false);
  const [state, setState] = useStorage({
    key: storageKey,
    storage: storageRef.current,
    preprocessValue(value) {
      const stateWithoutExpiredPendingApprovals = {
        activeIdentity: value.activeIdentity,
        identities: value.identities.filter(
          (signer) => signer.status !== "pending_approval" || signer.deadline && signer.deadline > Math.floor(Date.now() / 1e3)
        )
      };
      stateWithoutExpiredPendingApprovals.activeIdentity = stateWithoutExpiredPendingApprovals.identities.find(
        (identity) => identity._id === stateWithoutExpiredPendingApprovals.activeIdentity?._id
      ) ?? null;
      return stateWithoutExpiredPendingApprovals;
    },
    initialValue: {
      activeIdentity: null,
      identities: []
    }
  });
  const onImpersonateRef = useRef(onImpersonate);
  onImpersonateRef.current = onImpersonate;
  const onLogInRef = useRef(onLogIn);
  onLogInRef.current = onLogIn;
  const onLogInStartRef = useRef(onLogInStart);
  onLogInStartRef.current = onLogInStart;
  const onLogOutRef = useRef(onLogOut);
  onLogOutRef.current = onLogOut;
  const onIdentityRemoveRef = useRef(onIdentityRemove);
  onIdentityRemoveRef.current = onIdentityRemove;
  const onIdentitySelectRef = useRef(onIdentitySelect);
  onIdentitySelectRef.current = onIdentitySelect;
  const generateUserIdRef = useRef(generateUserId);
  generateUserIdRef.current = generateUserId;
  const createFarcasterSigner = useCallback(async () => {
    try {
      const keypair = await createKeypairEDDSA();
      const keypairString = convertKeypairToHex(keypair);
      const authorizationResponse = await fetch(
        // real signer or local one are handled by local route so we don't need to expose anything to client side bundle
        signerUrl,
        {
          method: "POST",
          body: JSON.stringify({
            publicKey: keypairString.publicKey
          })
        }
      );
      const authorizationBody = await authorizationResponse.json();
      if (authorizationResponse.status === 200) {
        const {
          signature,
          requestFid,
          deadline,
          requestSigner,
          sponsorship
        } = authorizationBody;
        const {
          result: { signedKeyRequest }
        } = await (await fetch(`https://api.warpcast.com/v2/signed-key-requests`, {
          method: "POST",
          headers: {
            "Content-Type": "application/json"
          },
          body: JSON.stringify({
            key: keypairString.publicKey,
            signature,
            requestFid,
            deadline,
            sponsorship
          })
        })).json();
        const deepLinkUrl = new URL(signedKeyRequest.deeplinkUrl);
        const signedKeyRequestToken = deepLinkUrl.searchParams.get("token");
        const signerApprovalUrl = new URL(
          "https://client.warpcast.com/deeplinks/signed-key-request"
        );
        if (!signedKeyRequestToken) {
          throw new Error("No token found in the deep link URL");
        }
        signerApprovalUrl.searchParams.set("token", signedKeyRequestToken);
        await setState((currentState) => {
          const newState = identityReducer(currentState, {
            type: "LOGIN_START",
            id: generateUserIdRef.current(),
            publicKey: keypairString.publicKey,
            privateKey: keypairString.privateKey,
            deadline,
            token: signedKeyRequestToken,
            signerApprovalUrl: signerApprovalUrl.toString(),
            requestFid: parseInt(requestFid, 10),
            requestSigner,
            signature
          });
          if (newState.activeIdentity?.status === "pending_approval") {
            onLogInStartRef.current?.(newState.activeIdentity);
          }
          return newState;
        });
        return {
          token: signedKeyRequestToken,
          signerApprovalUrl: signerApprovalUrl.toString()
        };
      } else if ("message" in authorizationBody) {
        throw new Error(authorizationBody.message);
      }
      throw new Error("Could not request signer approval.");
    } catch (error) {
      console.error("@frames.js/render: API Call failed", error);
      throw error;
    }
  }, [setState, signerUrl]);
  const impersonateUser = useCallback(
    async (fid) => {
      try {
        setIsLoading(true);
        const keypair = await createKeypairEDDSA();
        const { privateKey, publicKey } = convertKeypairToHex(keypair);
        await setState((currentState) => {
          const newState = identityReducer(currentState, {
            type: "IMPERSONATE",
            id: generateUserIdRef.current(),
            fid,
            privateKey,
            publicKey
          });
          if (newState.activeIdentity?.status === "impersonating") {
            onImpersonateRef.current?.(newState.activeIdentity);
          }
          return newState;
        });
      } finally {
        setIsLoading(false);
      }
    },
    [setState]
  );
  const onSignerlessFramePress = useCallback(() => {
    onMissingIdentity();
    return Promise.resolve();
  }, [onMissingIdentity]);
  const createSigner = useCallback(async () => {
    setIsLoading(true);
    const result = await createFarcasterSigner();
    setIsLoading(false);
    return result;
  }, [createFarcasterSigner]);
  const logout = useCallback(async () => {
    await setState((currentState) => {
      if (currentState.activeIdentity) {
        onLogOutRef.current?.(currentState.activeIdentity);
      }
      return identityReducer(currentState, { type: "LOGOUT" });
    });
  }, [setState]);
  const removeIdentity = useCallback(async () => {
    await setState((currentState) => {
      if (currentState.activeIdentity) {
        onIdentityRemoveRef.current?.(currentState.activeIdentity);
      }
      return identityReducer(currentState, { type: "REMOVE" });
    });
  }, [setState]);
  const farcasterUser = state.activeIdentity;
  const visibilityDetector = visibilityChangeDetectionHook();
  useEffect(() => {
    if (farcasterUser && farcasterUser.status === "pending_approval") {
      const startPolling = () => {
        identityPoller.start(farcasterUser.token).then((signedKeyRequest) => {
          if (signedKeyRequest) {
            return setState((currentState) => {
              const newState = identityReducer(currentState, {
                type: "LOGIN_SUCCESS",
                signedKeyRequest
              });
              if (newState.activeIdentity?.status === "approved") {
                onLogInRef.current?.(newState.activeIdentity);
              }
              return newState;
            });
          }
        }).catch((e) => {
          console.error(
            "Error while polling for the signer approval status",
            e
          );
        });
      };
      const unregisterVisibilityChangeListener = visibilityDetector.register(
        (visible) => {
          if (visible) {
            startPolling();
          } else {
            identityPoller.stop();
          }
        }
      );
      startPolling();
      return () => {
        identityPoller.stop();
        unregisterVisibilityChangeListener();
      };
    }
  }, [farcasterUser, identityPoller, visibilityDetector, setState]);
  const selectIdentity = useCallback(
    async (id) => {
      await setState((currentState) => {
        const newState = identityReducer(currentState, {
          type: "SELECT_IDENTITY",
          id
        });
        if (newState.activeIdentity) {
          onIdentitySelectRef.current?.(newState.activeIdentity);
        }
        return newState;
      });
    },
    [setState]
  );
  return useMemo(
    () => ({
      signer: farcasterUser,
      hasSigner: farcasterUser?.status === "approved" || farcasterUser?.status === "impersonating",
      signFrameAction,
      isLoadingSigner: isLoading,
      impersonateUser,
      onSignerlessFramePress,
      createSigner,
      logout,
      removeIdentity,
      identities: state.identities,
      selectIdentity,
      identityPoller
    }),
    [
      farcasterUser,
      identityPoller,
      impersonateUser,
      isLoading,
      logout,
      createSigner,
      onSignerlessFramePress,
      removeIdentity,
      selectIdentity,
      state.identities
    ]
  );
}

export {
  useFarcasterMultiIdentity
};
